@@ -1,5 +1,5 @@
 /**
- *  Copyright 2009-2012 10gen, Inc.
+ *  Copyright 2009-2013 10gen, Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -24,20 +24,6 @@
 #include "bson_helpers.h"
 #include "mini_bson.h"
 
-#ifdef WIN32
-#include <winsock2.h>
-#include <ws2tcpip.h>
-#else
-#include <netinet/in.h>
-#include <netinet/tcp.h>
-#include <fcntl.h>
-#include <netdb.h>
-#include <sys/un.h>
-#include <sys/socket.h>
-#include <unistd.h>
-#include <sys/time.h>
-#endif
-
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -48,6 +34,8 @@
 
 #define MONGO_REPLY_FLAG_QUERY_FAILURE 0x02
 
+static void mongo_close_socket(int socket, int why);
+
 /* Helper functions */
 int mongo_connection_get_reqid(mongo_connection *con)
 {
@@ -87,8 +75,12 @@ static int mongo_util_connect__sockaddr(struct sockaddr *sa, int family, char *h
 	return 1;
 }
 
-/* This function does the actual connecting */
-int mongo_connection_connect(char *host, int port, int timeout, char **error_message)
+/* This function does the actual connecting The results of this function are
+ * stored in mongo_connection->socket, which is a void* to be able to store
+ * various different backends (f.e. the PHP io_streams stores a php_stream*)
+ *
+ * Returns an integer (masquerading as a void*) on success, NULL on failure. */
+void* mongo_connection_connect(mongo_con_manager *manager, mongo_server_def *server, mongo_server_options *options, char **error_message)
 {
 	struct sockaddr*   sa;
 	struct sockaddr_in si;
@@ -122,19 +114,19 @@ int mongo_connection_connect(char *host, int port, int timeout, char **error_mes
 	error = WSAStartup(version, &wsaData);
 
 	if (error != 0) {
-		return -1;
+		return NULL;
 	}
 
 	/* create socket */
 	tmp_socket = socket(family, SOCK_STREAM, 0);
 	if (tmp_socket == INVALID_SOCKET) {
 		*error_message = strdup(strerror(errno));
-		return -1;
+		return NULL;
 	}
 
 #else
 	/* domain socket */
-	if (port == 0) {
+	if (server->port == 0) {
 		family = AF_UNIX;
 		sa = (struct sockaddr*)(&su);
 		sn = sizeof(su);
@@ -147,17 +139,17 @@ int mongo_connection_connect(char *host, int port, int timeout, char **error_mes
 	/* create socket */
 	if ((tmp_socket = socket(family, SOCK_STREAM, 0)) == -1) {
 		*error_message = strdup(strerror(errno));
-		return -1;
+		return NULL;
 	}
 #endif
 
 	/* TODO: Move this to within the loop & use real timeout setting */
 	/* connection timeout: set in ms (current default 1 sec) */
-	tval.tv_sec = timeout <= 0 ? 1 : timeout / 1000;
-	tval.tv_usec = timeout <= 0 ? 0 : (timeout % 1000) * 1000;
+	tval.tv_sec = options->connectTimeoutMS <= 0 ? 1 : options->connectTimeoutMS / 1000;
+	tval.tv_usec = options->connectTimeoutMS <= 0 ? 0 : (options->connectTimeoutMS % 1000) * 1000;
 
 	/* get addresses */
-	if (mongo_util_connect__sockaddr(sa, family, host, port, error_message) == 0) {
+	if (mongo_util_connect__sockaddr(sa, family, server->host, server->port, error_message) == 0) {
 		goto error;
 	}
 
@@ -195,7 +187,7 @@ int mongo_connection_connect(char *host, int port, int timeout, char **error_mes
 
 			if (select(tmp_socket+1, &rset, &wset, &eset, &tval) == 0) {
 				*error_message = malloc(256);
-				snprintf(*error_message, 256, "Timed out after %d ms", timeout);
+				snprintf(*error_message, 256, "Timed out after %d ms", options->connectTimeoutMS);
 				goto error;
 			}
 
@@ -226,21 +218,32 @@ int mongo_connection_connect(char *host, int port, int timeout, char **error_mes
 #else
 	fcntl(tmp_socket, F_SETFL, FLAGS);
 #endif
-	return tmp_socket;
+
+	return (void *) (long) tmp_socket;
 
 error:
-#ifdef WIN32
-	shutdown((tmp_socket), 2);
-	closesocket(tmp_socket);
-	WSACleanup();
-#else
-	shutdown((tmp_socket), 2);
-	close(tmp_socket);
-#endif
-	return -1;
+	mongo_close_socket(tmp_socket, MONGO_CLOSE_BROKEN);
+	return NULL;
 }
 
-mongo_connection *mongo_connection_create(mongo_con_manager *manager, mongo_server_def *server_def, mongo_server_options *options, char **error_message)
+static void mongo_close_socket(int socket, int why)
+{
+#ifdef WIN32
+	shutdown(socket, SD_BOTH);
+	closesocket(socket);
+	WSACleanup();
+#else
+	shutdown(socket, SHUT_RDWR);
+	close(socket);
+#endif
+}
+
+void mongo_connection_close(mongo_connection *con, int why)
+{
+	mongo_close_socket((int) (long) con->socket, why);
+}
+
+mongo_connection *mongo_connection_create(mongo_con_manager *manager, char *hash, mongo_server_def *server_def, mongo_server_options *options, char **error_message)
 {
 	mongo_connection *tmp;
 
@@ -250,92 +253,106 @@ mongo_connection *mongo_connection_create(mongo_con_manager *manager, mongo_serv
 	tmp->last_reqid = rand();
 	tmp->connection_type = MONGO_NODE_STANDALONE;
 
+	/* Store hash */
+	tmp->hash = strdup(hash);
+
 	/* Connect */
 	mongo_manager_log(manager, MLOG_CON, MLOG_INFO, "connection_create: creating new connection for %s:%d", server_def->host, server_def->port);
-	tmp->socket = mongo_connection_connect(server_def->host, server_def->port, options->connectTimeoutMS, error_message);
-	if (tmp->socket == -1) {
+	tmp->socket = manager->connect(manager, server_def, options, error_message);
+	if (!tmp->socket) {
 		mongo_manager_log(manager, MLOG_CON, MLOG_WARN, "connection_create: error while creating connection for %s:%d: %s", server_def->host, server_def->port, *error_message);
+		mongo_manager_blacklist_register(manager, tmp);
+		free(tmp->hash);
 		free(tmp);
 		return NULL;
 	}
 
 	/* We call get_server_flags to the maxBsonObjectSize data */
-	mongo_connection_get_server_flags(manager, tmp, options, (char**) &error_message);
+	if (!mongo_connection_get_server_flags(manager, tmp, options, error_message)) {
+		mongo_manager_log(manager, MLOG_CON, MLOG_WARN, "server_flags: error while getting the server configuration %s:%d: %s", server_def->host, server_def->port, *error_message);
+		free(tmp);
+		return NULL;
+	}
 
 	return tmp;
 }
 
-void mongo_connection_destroy(mongo_con_manager *manager, mongo_connection *con)
+void mongo_connection_destroy(mongo_con_manager *manager, void *data, int why)
 {
 	int current_pid, connection_pid;
 	int i;
+	mongo_connection *con = (mongo_connection *)data;
 
 	current_pid = getpid();
 	connection_pid = mongo_server_hash_to_pid(con->hash);
 
 	/* Only close the connection if it matches the current PID */
-	if (current_pid != connection_pid) {
-		mongo_manager_log(manager, MLOG_CON, MLOG_INFO, "mongo_connection_destroy: The process pid (%d) for %s doesn't match the connection pid (%d).", current_pid, con->hash, connection_pid);
+	if (current_pid == connection_pid) {
+		mongo_manager_log(manager, MLOG_CON, MLOG_FINE, "mongo_connection_destroy: Destroying connection object for %s", con->hash);
+
+		if (con->socket) {
+			mongo_manager_log(manager, MLOG_CON, MLOG_FINE, "mongo_connection_destroy: Closing socket for %s.", con->hash);
+			manager->close(con, why);
+			con->socket = NULL;
+
+			for (i = 0; i < con->tag_count; i++) {
+				free(con->tags[i]);
+			}
+			free(con->tags);
+
+			if (con->cleanup_list) {
+				mongo_connection_deregister_callback *ptr = con->cleanup_list;
+				mongo_connection_deregister_callback *prev;
+				do {
+					if (ptr->callback_data) {
+						ptr->mongo_cleanup_cb(ptr->callback_data);
+					}
+
+					if (!ptr->next) {
+						free(ptr);
+						ptr = NULL;
+						break;
+					}
+					prev = ptr;
+					ptr = ptr->next;
+					free(prev);
+					prev = NULL;
+				} while(1);
+				con->cleanup_list = NULL;
+			}
+			free(con->hash);
+			free(con);
+		}
 	} else {
-		mongo_manager_log(manager, MLOG_CON, MLOG_FINE, "mongo_connection_destroy: Closing socket for %s.", con->hash);
-
-#ifdef WIN32
-		shutdown(con->socket, SD_BOTH);
-		closesocket(con->socket);
-		WSACleanup();
-#else
-		shutdown(con->socket, SHUT_RDWR);
-		close(con->socket);
-#endif
-		for (i = 0; i < con->tag_count; i++) {
-			free(con->tags[i]);
-		}
-		if (con->cleanup_list) {
-			mongo_connection_deregister_callback *ptr = con->cleanup_list;
-			mongo_connection_deregister_callback *prev;
-			do {
-				if (ptr->callback_data) {
-					ptr->mongo_cleanup_cb(ptr->callback_data);
-				}
-
-				if (!ptr->next) {
-					free(ptr);
-					ptr = NULL;
-					break;
-				}
-				prev = ptr;
-				ptr = ptr->next;
-				free(prev);
-				prev = NULL;
-			} while(1);
-			con->cleanup_list = NULL;
-		}
-		free(con->tags);
-		free(con->hash);
-		free(con);
+		mongo_manager_log(manager, MLOG_CON, MLOG_INFO, "mongo_connection_destroy: The process pid (%d) for %s doesn't match the connection pid (%d).", current_pid, con->hash, connection_pid);
 	}
 }
 
+void mongo_connection_forget(mongo_con_manager *manager, mongo_connection *con)
+{
+	/* FIXME: Should we remove it from our connection registry ? */
+}
+
 #define MONGO_REPLY_HEADER_SIZE 36
 
 /* Returns 1 if it worked, and 0 if it didn't. If 0 is returned, *error_message
- * is set and must be freed */
+ * is set and must be free()d. On success *data_buffer is set and must be free()d */
 static int mongo_connect_send_packet(mongo_con_manager *manager, mongo_connection *con, mongo_server_options *options, mcon_str *packet, char **data_buffer, char **error_message)
 {
 	int            read;
 	uint32_t       data_size;
 	char           reply_buffer[MONGO_REPLY_HEADER_SIZE];
 	uint32_t       flags; /* To check for query reply status */
-	char          *recv_error_message;
 
 	/* Send and wait for reply */
-	mongo_io_send(con->socket, packet->d, packet->l, error_message);
+	if (manager->send(con, options, packet->d, packet->l, error_message) == -1) {
+		mcon_str_ptr_dtor(packet);
+		return 0;
+	}
 	mcon_str_ptr_dtor(packet);
-	read = mongo_io_recv_header(con->socket, options, reply_buffer, MONGO_REPLY_HEADER_SIZE, &recv_error_message);
-	if (read == -1) {
-		*error_message = malloc(256);
-		snprintf(*error_message, 256, "send_package: error reading from socket: %s", recv_error_message);
-		free(recv_error_message);
+	read = manager->recv_header(con, options, options->socketTimeoutMS, reply_buffer, MONGO_REPLY_HEADER_SIZE, error_message);
+	if (read < 0) {
+		/* Error already populated */
 		return 0;
 	}
 
@@ -362,7 +379,8 @@ static int mongo_connect_send_packet(mongo_con_manager *manager, mongo_connectio
 
 	/* Read data */
 	*data_buffer = malloc(data_size + 1);
-	if (!mongo_io_recv_data(con->socket, options, *data_buffer, data_size, error_message)) {
+	if (manager->recv_data(con, options, options->socketTimeoutMS, *data_buffer, data_size, error_message) <= 0) {
+		free(*data_buffer);
 		return 0;
 	}
 
@@ -385,29 +403,36 @@ static int mongo_connect_send_packet(mongo_con_manager *manager, mongo_connectio
 			*error_message = strdup("send_package: the query returned an unknown error");
 		}
 
+		free(*data_buffer);
 		return 0;
 	}
 
 	return 1;
 }
 
-/**
- * Sends a ping command to the server and stores the result.
+int mongo_connection_ping_check(mongo_con_manager *manager, int last_ping, struct timeval *start)
+{
+	gettimeofday(start, NULL);
+	if ((last_ping + manager->ping_interval) > start->tv_sec) {
+		mongo_manager_log(manager, MLOG_CON, MLOG_FINE, "is_ping: skipping: last ran at %ld, now: %ld, time left: %ld", last_ping, start->tv_sec, last_ping + manager->ping_interval - start->tv_sec);
+		return 0;
+	}
+	return 1;
+}
+
+/* Sends a ping command to the server and stores the result.
  *
- * Returns 1 when it worked, and 0 when an error was encountered.
- */
+ * Returns 1 when it worked, and 0 when an error was encountered. */
 int mongo_connection_ping(mongo_con_manager *manager, mongo_connection *con, mongo_server_options *options, char **error_message)
 {
 	mcon_str      *packet;
 	struct timeval start, end;
 	char          *data_buffer;
 
-	gettimeofday(&start, NULL);
-	if ((con->last_ping + manager->ping_interval) > start.tv_sec) {
-		mongo_manager_log(manager, MLOG_CON, MLOG_FINE, "is_ping: skipping: last ran at %ld, now: %ld, time left: %ld", con->last_ping, start.tv_sec, con->last_ping + manager->ping_interval - start.tv_sec);
-		return 2;
+	/* If we haven't hit the ping_interval yet, then there is no need to do a roundtrip to the server */
+	if (!mongo_connection_ping_check(manager, con->last_ping, &start)) {
+		return 1;
 	}
-
 	mongo_manager_log(manager, MLOG_CON, MLOG_INFO, "is_ping: pinging %s", con->hash);
 	packet = bson_create_ping_packet(con);
 	if (!mongo_connect_send_packet(manager, con, options, packet, &data_buffer, error_message)) {
@@ -427,8 +452,7 @@ int mongo_connection_ping(mongo_con_manager *manager, mongo_connection *con, mon
 	return 1;
 }
 
-/**
- * Sends an ismaster command to the server and returns an array of new
+/* Sends an ismaster command to the server and returns an array of new
  * connectable nodes
  *
  * Returns:
@@ -437,14 +461,13 @@ int mongo_connection_ping(mongo_con_manager *manager, mongo_connection *con, mon
  * 2: when is master wasn't run due to the time-out limit
  * 3: when it all worked, but we need to remove the seed host (due to its name
  *    not being what the server thought it is) - in that case, the server in
- *    the last argument is changed
- */
+ *    the last argument is changed */
 int mongo_connection_ismaster(mongo_con_manager *manager, mongo_connection *con, mongo_server_options *options, char **repl_set_name, int *nr_hosts, char ***found_hosts, char **error_message, mongo_server_def *server)
 {
 	mcon_str      *packet;
 	char          *data_buffer;
 	char          *set = NULL;      /* For replicaset in return */
-	char          *hosts, *ptr, *string;
+	char          *hosts, *passives = NULL, *ptr, *string;
 	unsigned char  ismaster = 0, secondary = 0, arbiter = 0;
 	char          *connected_name, *we_think_we_are;
 	struct timeval now;
@@ -468,35 +491,25 @@ int mongo_connection_ismaster(mongo_con_manager *manager, mongo_connection *con,
 
 	/* We find out whether the machine we connected too, is actually the
 	 * one we thought we were connecting too */
-	if (!bson_find_field_as_string(ptr, "me", &connected_name)) {
-		struct mcon_str *tmp;
-
-		mcon_str_ptr_init(tmp);
-		mcon_str_add(tmp, "Host does not seem to be a replicaset member (", 0);
-		mcon_str_add(tmp, mongo_server_hash_to_server(con->hash), 1);
-		mcon_str_add(tmp, ")", 0);
-
-		*error_message = strdup(tmp->d);
-		mcon_str_ptr_dtor(tmp);
-
-		mongo_manager_log(manager, MLOG_CON, MLOG_WARN, *error_message);
-		free(data_buffer);
-		return 0;
-	}
-
-	we_think_we_are = mongo_server_hash_to_server(con->hash);
-	if (strcmp(connected_name, we_think_we_are) == 0) {
-		mongo_manager_log(manager, MLOG_CON, MLOG_FINE, "ismaster: the server name matches what we thought it'd be (%s).", we_think_we_are);
+	/* MongoDB 1.8.x doesn't have the "me" field.
+	 * The replicaset verification is done next step (setName). */
+	if (bson_find_field_as_string(ptr, "me", &connected_name)) {
+		we_think_we_are = mongo_server_hash_to_server(con->hash);
+		if (strcmp(connected_name, we_think_we_are) == 0) {
+			mongo_manager_log(manager, MLOG_CON, MLOG_FINE, "ismaster: the server name matches what we thought it'd be (%s).", we_think_we_are);
+		} else {
+			mongo_manager_log(manager, MLOG_CON, MLOG_WARN, "ismaster: the server name (%s) did not match with what we thought it'd be (%s).", connected_name, we_think_we_are);
+			/* We reset the name as the server responded with a different name than
+			 * what we thought it was */
+			free(server->host);
+			server->host = mcon_strndup(connected_name, strchr(connected_name, ':') - connected_name);
+			server->port = atoi(strchr(connected_name, ':') + 1);
+			retval = 3;
+		}
+		free(we_think_we_are);
 	} else {
-		mongo_manager_log(manager, MLOG_CON, MLOG_WARN, "ismaster: the server name (%s) did not match with what we thought it'd be (%s).", connected_name, we_think_we_are);
-		/* We reset the name as the server responded with a different name than
-		 * what we thought it was */
-		free(server->host);
-		server->host = mcon_strndup(connected_name, strchr(connected_name, ':') - connected_name);
-		server->port = atoi(strchr(connected_name, ':') + 1);
-		retval = 3;
+		mongo_manager_log(manager, MLOG_CON, MLOG_WARN, "Can't find 'me' in ismaster response, possibly not a replicaset (%s)", mongo_server_hash_to_server(con->hash));
 	}
-	free(we_think_we_are);
 
 	/* Do replica set name test */
 	bson_find_field_as_string(ptr, "setName", &set);
@@ -545,11 +558,14 @@ int mongo_connection_ismaster(mongo_con_manager *manager, mongo_connection *con,
 	bson_find_field_as_bool(ptr, "secondary", &secondary);
 	bson_find_field_as_bool(ptr, "arbiterOnly", &arbiter);
 
+	mongo_manager_log(manager, MLOG_CON, MLOG_INFO, "ismaster: set name: %s, ismaster: %d, secondary: %d, is_arbiter: %d", set, ismaster, secondary, arbiter);
+
 	/* Find all hosts */
 	bson_find_field_as_array(ptr, "hosts", &hosts);
-	mongo_manager_log(manager, MLOG_CON, MLOG_INFO, "ismaster: set name: %s, ismaster: %d, secondary: %d, is_arbiter: %d", set, ismaster, secondary, arbiter);
+	bson_find_field_as_array(ptr, "passives", &passives);
 	*nr_hosts = 0;
 
+	/* Iterate over the "hosts" document */
 	ptr = hosts;
 	while (bson_array_find_next_string(&ptr, NULL, &string)) {
 		(*nr_hosts)++;
@@ -558,6 +574,17 @@ int mongo_connection_ismaster(mongo_con_manager *manager, mongo_connection *con,
 		mongo_manager_log(manager, MLOG_CON, MLOG_INFO, "found host: %s", string);
 	}
 
+	/* Iterate over the "passives" document (priority=0) */
+	if (passives) {
+		ptr = passives;
+		while (bson_array_find_next_string(&ptr, NULL, &string)) {
+			(*nr_hosts)++;
+			*found_hosts = realloc(*found_hosts, (*nr_hosts) * sizeof(char*));
+			(*found_hosts)[*nr_hosts-1] = strdup(string);
+			mongo_manager_log(manager, MLOG_CON, MLOG_INFO, "found host: %s (passive)", string);
+		}
+	}
+
 	/* Set connection type depending on flags */
 	if (ismaster) {
 		con->connection_type = MONGO_NODE_PRIMARY;
@@ -577,15 +604,13 @@ int mongo_connection_ismaster(mongo_con_manager *manager, mongo_connection *con,
 	return retval;
 }
 
-/**
- * Sends an ismaster command to the server to find server flags
+/* Sends an ismaster command to the server to find server flags
  *
- * Returns 1 when it worked, and 0 when an error was encountered.
- */
+ * Returns 1 when it worked, and 0 when an error was encountered. */
 int mongo_connection_get_server_flags(mongo_con_manager *manager, mongo_connection *con, mongo_server_options *options, char **error_message)
 {
 	mcon_str      *packet;
-	int32_t        max_bson_size = 0;
+	int32_t        max_bson_size = 0, max_message_size = 0;
 	char          *data_buffer;
 	char          *ptr;
 	char          *tags;
@@ -612,6 +637,17 @@ int mongo_connection_get_server_flags(mongo_con_manager *manager, mongo_connecti
 		mongo_manager_log(manager, MLOG_CON, MLOG_FINE, "get_server_flags: can't find maxBsonObjectSize, defaulting to %d", con->max_bson_size);
 	}
 
+	/* Find max message size */
+	if (bson_find_field_as_int32(ptr, "maxMessageSizeBytes", &max_message_size)) {
+		mongo_manager_log(manager, MLOG_CON, MLOG_FINE, "get_server_flags: setting maxMessageSizeBytes to %d", max_message_size);
+		con->max_message_size = max_message_size;
+	} else {
+		/* This seems to be a pre-2.4 MongoDB installation, where we need to
+		 * default to two times the max BSON size */
+		con->max_message_size = 2 * con->max_bson_size;
+		mongo_manager_log(manager, MLOG_CON, MLOG_FINE, "get_server_flags: can't find maxMessageSizeBytes, defaulting to %d", con->max_message_size);
+	}
+
 	/* Find msg and whether it contains "isdbgrid" */
 	if (bson_find_field_as_string(ptr, "msg", (char**) &msg)) {
 		if (strcmp(msg, "isdbgrid") == 0) {
@@ -645,11 +681,9 @@ int mongo_connection_get_server_flags(mongo_con_manager *manager, mongo_connecti
 	return 1;
 }
 
-/**
- * Sends a getnonce command to the server for authentication
+/* Sends a getnonce command to the server for authentication
  *
- * Returns the nonsense when it worked, or NULL if it didn't.
- */
+ * Returns the nonsense when it worked, or NULL if it didn't. */
 char *mongo_connection_getnonce(mongo_con_manager *manager, mongo_connection *con, mongo_server_options *options, char **error_message)
 {
 	mcon_str      *packet;
@@ -684,11 +718,9 @@ char *mongo_connection_getnonce(mongo_con_manager *manager, mongo_connection *co
 	return retval;
 }
 
-/**
- * Authenticates a connection
+/* Authenticates a connection
  *
- * Returns 1 when it worked, or 0 when it didn't - with the error_message set.
- */
+ * Returns 1 when it worked, or 0 when it didn't - with the error_message set. */
 int mongo_connection_authenticate(mongo_con_manager *manager, mongo_connection *con, mongo_server_options *options, char *database, char *username, char *password, char *nonce, char **error_message)
 {
 	mcon_str      *packet;
@@ -723,7 +755,6 @@ int mongo_connection_authenticate(mongo_con_manager *manager, mongo_connection *
 	free(key);
 
 	if (!mongo_connect_send_packet(manager, con, options, packet, &data_buffer, error_message)) {
-		free(data_buffer);
 		return 0;
 	}
 
@@ -749,3 +780,12 @@ int mongo_connection_authenticate(mongo_con_manager *manager, mongo_connection *
 
 	return 1;
 }
+
+/*
+ * Local variables:
+ * tab-width: 4
+ * c-basic-offset: 4
+ * End:
+ * vim600: fdm=marker
+ * vim: noet sw=4 ts=4
+ */
