@@ -1,5 +1,5 @@
 /**
- *  Copyright 2009-2012 10gen, Inc.
+ *  Copyright 2009-2013 10gen, Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -48,7 +48,7 @@ static void mongo_print_connection_info(mongo_con_manager *manager, mongo_connec
 	mongo_manager_log(manager, MLOG_RS, level,
 		"- connection: type: %s, socket: %d, ping: %d, hash: %s",
 		mongo_connection_type(con->connection_type),
-		con->socket,
+		42, /* FIXME: STREAMS: Maybe we do need a union here: con->socket, */
 		con->ping_ms,
 		con->hash
 	);
@@ -90,13 +90,15 @@ static mcon_collection *filter_connections(mongo_con_manager *manager, int types
 
 	mongo_manager_log(manager, MLOG_RS, MLOG_FINE, "filter_connections: adding connections:");
 	while (ptr) {
-		connection_pid = mongo_server_hash_to_pid(ptr->connection->hash);
+		mongo_connection *con = (mongo_connection *) ptr->data;
+		connection_pid = mongo_server_hash_to_pid(con->hash);
 
 		if (connection_pid != current_pid) {
-			mongo_manager_log(manager, MLOG_RS, MLOG_FINE, "filter_connections: skipping %s as it doesn't match the current pid (%d)", ptr->connection->hash, current_pid);
-		} else if (ptr->connection->connection_type & types) {
-			mongo_print_connection_info(manager, ptr->connection, MLOG_FINE);
-			mcon_collection_add(col, ptr->connection);
+			mongo_manager_log(manager, MLOG_RS, MLOG_FINE, "filter_connections: skipping %s as it doesn't match the current pid (%d)", con->hash, current_pid);
+			manager->forget(manager, con);
+		} else if (con->connection_type & types) {
+			mongo_print_connection_info(manager, con, MLOG_FINE);
+			mcon_collection_add(col, con);
 		}
 		ptr = ptr->next;
 	}
@@ -620,3 +622,12 @@ void mongo_read_preference_replace(mongo_read_preference *from, mongo_read_prefe
 	mongo_read_preference_dtor(to);
 	mongo_read_preference_copy(from, to);
 }
+
+/*
+ * Local variables:
+ * tab-width: 4
+ * c-basic-offset: 4
+ * End:
+ * vim600: fdm=marker
+ * vim: noet sw=4 ts=4
+ */
