@@ -1,5 +1,5 @@
 /**
- *  Copyright 2009-2012 10gen, Inc.
+ *  Copyright 2009-2013 10gen, Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -18,18 +18,18 @@
 #include <errno.h>
 
 #ifndef WIN32
-#include <sys/types.h>
+# include <sys/types.h>
 #endif
 
 #include <php.h>
 #include <zend_exceptions.h>
 #include "ext/standard/php_smart_str.h"
+#include "ext/standard/file.h"
 
 #include "php_mongo.h"
 #include "mongoclient.h"
 #include "db.h"
 #include "cursor.h"
-#include "mongo_types.h"
 #include "bson.h"
 
 #include "util/log.h"
@@ -49,14 +49,12 @@ static int close_connection(mongo_con_manager *manager, mongo_connection *connec
 zend_object_handlers mongo_default_handlers;
 zend_object_handlers mongoclient_handlers;
 
-ZEND_EXTERN_MODULE_GLOBALS(mongo);
+ZEND_EXTERN_MODULE_GLOBALS(mongo)
 
 zend_class_entry *mongo_ce_MongoClient;
 
-extern zend_class_entry *mongo_ce_DB,
-  *mongo_ce_Cursor,
-  *mongo_ce_Exception,
-  *mongo_ce_ConnectionException;
+extern zend_class_entry *mongo_ce_DB, *mongo_ce_Cursor, *mongo_ce_Exception;
+extern zend_class_entry *mongo_ce_ConnectionException;
 
 MONGO_ARGINFO_STATIC ZEND_BEGIN_ARG_INFO_EX(arginfo___construct, 0, ZEND_RETURN_VALUE, 0)
 	ZEND_ARG_INFO(0, server)
@@ -103,7 +101,7 @@ static zend_function_entry mongo_methods[] = {
 	PHP_ME(MongoClient, getHosts, arginfo_no_parameters, ZEND_ACC_PUBLIC)
 	PHP_ME(MongoClient, close, arginfo_no_parameters, ZEND_ACC_PUBLIC)
 
-{ NULL, NULL, NULL }
+	{ NULL, NULL, NULL }
 };
 
 /* {{{ php_mongoclient_free
@@ -274,7 +272,7 @@ void mongo_init_MongoClient(TSRMLS_D)
 }
 
 /* {{{ Helper for connecting the servers */
-mongo_connection *php_mongo_connect(mongoclient *link TSRMLS_DC)
+mongo_connection *php_mongo_connect(mongoclient *link, int flags TSRMLS_DC)
 {
 	mongo_connection *con;
 	char *error_message = NULL;
@@ -282,17 +280,18 @@ mongo_connection *php_mongo_connect(mongoclient *link TSRMLS_DC)
 	/* We don't care about the result so although we assign it to a var, we
 	 * only do that to handle errors and return it so that the calling function
 	 * knows whether a connection could be obtained or not. */
-	con = mongo_get_read_write_connection(link->manager, link->servers, MONGO_CON_FLAG_READ, (char **) &error_message);
-	if (!con) {
-		if (error_message) {
-			zend_throw_exception(mongo_ce_ConnectionException, error_message, 71 TSRMLS_CC);
-			free(error_message);
-		} else {
-			zend_throw_exception(mongo_ce_ConnectionException, "Unknown error obtaining connection", 72 TSRMLS_CC);
-		}
-		return NULL;
+	con = mongo_get_read_write_connection(link->manager, link->servers, flags, (char **) &error_message);
+	if (con) {
+		return con;
 	}
-	return con;
+
+	if (error_message) {
+		zend_throw_exception(mongo_ce_ConnectionException, error_message, 71 TSRMLS_CC);
+		free(error_message);
+	} else {
+		zend_throw_exception(mongo_ce_ConnectionException, "Unknown error obtaining connection", 72 TSRMLS_CC);
+	}
+	return NULL;
 }
 /* }}} */
 
@@ -319,7 +318,7 @@ int mongo_store_option_wrapper(mongo_con_manager *manager, mongo_servers *server
 			zend_hash_move_forward_ex(Z_ARRVAL_PP(option_value), &pos)
 		) {
 			convert_to_string_ex(opt_entry);
-			error = mongo_store_option(manager, servers, option_name, Z_STRVAL_PP(opt_entry), (char **) &error_message);
+			error = mongo_store_option(manager, servers, option_name, Z_STRVAL_PP(opt_entry), error_message);
 			if (error) {
 				return error;
 			}
@@ -332,8 +331,10 @@ int mongo_store_option_wrapper(mongo_con_manager *manager, mongo_servers *server
 }
 /* }}} */
 
-/* {{{ MongoClient->__construct
- */
+/* {{{ proto MongoClient MongoClient->__construct([string connection_string [, array mongo_options [, array driver_options]]])
+   Creates a new MongoClient object for mongo[d|s]. mongo_options are the same
+   options as the connection_string, while driver_options is additional PHP
+   MongoDB options, like stream context and callbacks. */
 PHP_METHOD(MongoClient, __construct)
 {
 	php_mongo_ctor(INTERNAL_FUNCTION_PARAM_PASSTHRU, 0);
@@ -346,6 +347,7 @@ void php_mongo_ctor(INTERNAL_FUNCTION_PARAMETERS, int bc)
 	zend_bool     connect = 1;
 	zval         *options = 0;
 	zval         *slave_okay = 0;
+	zval         *zdoptions = NULL;
 	mongoclient  *link;
 	zval        **opt_entry;
 	char         *opt_key;
@@ -355,7 +357,7 @@ void php_mongo_ctor(INTERNAL_FUNCTION_PARAMETERS, int bc)
 	ulong         num_key;
 	HashPosition  pos;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|s!a!/", &server, &server_len, &options) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|s!a!/a!/", &server, &server_len, &options, &zdoptions) == FAILURE) {
 		zval *object = getThis();
 		ZVAL_NULL(object);
 		return;
@@ -411,9 +413,17 @@ void php_mongo_ctor(INTERNAL_FUNCTION_PARAMETERS, int bc)
 			switch (zend_hash_get_current_key_ex(Z_ARRVAL_P(options), &opt_key, &opt_key_len, &num_key, 0, &pos)) {
 				case HASH_KEY_IS_STRING: {
 					int error = 0;
+
 					error = mongo_store_option_wrapper(link->manager, link->servers, opt_key, opt_entry, (char **)&error_message);
 
 					switch (error) {
+						case -1: /* Deprecated options */
+							if (strcasecmp(opt_key, "slaveOkay") == 0) {
+								php_error_docref(NULL TSRMLS_CC, MONGO_E_DEPRECATED, "The 'slaveOkay' option is deprecated. Please switch to read-preferences");
+							} else if (strcasecmp(opt_key, "timeout") == 0) {
+								php_error_docref(NULL TSRMLS_CC, MONGO_E_DEPRECATED, "The 'timeout' option is deprecated. Please use 'connectTimeoutMS' instead");
+							}
+							break;
 						case 4: /* Special options parameters, invalid for URL parsing - only possiblity is 'connect' for now */
 							if (strcasecmp(opt_key, "connect") == 0) {
 								convert_to_boolean_ex(opt_entry);
@@ -439,10 +449,32 @@ void php_mongo_ctor(INTERNAL_FUNCTION_PARAMETERS, int bc)
 		}
 	}
 
+#if !MONGO_PHP_STREAMS
+	if (link->servers->options.ssl) {
+		zend_throw_exception(mongo_ce_ConnectionException, "SSL support is only available when compiled against PHP Streams", 26 TSRMLS_CC);
+		return;
+	}
+	if (zdoptions) {
+		zend_throw_exception(mongo_ce_ConnectionException, "Driver options are only available when compiled against PHP Streams", 27 TSRMLS_CC);
+		return;
+	}
+#endif
+	if (zdoptions) {
+		/* Possibly add more indexes in the future, like "log_queries",
+		 * "log_commands", "log_failures", "log_metadata"... */
+		zval **zcontext;
+
+		if (zend_hash_find(Z_ARRVAL_P(zdoptions), "context", strlen("context") + 1, (void**)&zcontext) == SUCCESS) {
+			link->servers->options.ctx = php_stream_context_from_zval(*zcontext, PHP_FILE_NO_DEFAULT_CONTEXT);
+			mongo_manager_log(link->manager, MLOG_CON, MLOG_INFO, "Found Stream context");
+		}
+	}
+
+
 	slave_okay = zend_read_static_property(mongo_ce_Cursor, "slaveOkay", strlen("slaveOkay"), NOISY TSRMLS_CC);
 	if (Z_BVAL_P(slave_okay)) {
 		if (link->servers->read_pref.type != MONGO_RP_PRIMARY) {
-			/* the server already has read preferences configured, but we're still
+			/* The server already has read preferences configured, but we're still
 			 * trying to set slave okay. The spec says that's an error, so we
 			 * throw an exception with code 23 (defined in php_mongo.h) */
 			zend_throw_exception(mongo_ce_ConnectionException, "You can not use both slaveOkay and read-preferences. Please switch to read-preferences.", 23 TSRMLS_CC);
@@ -455,7 +487,7 @@ void php_mongo_ctor(INTERNAL_FUNCTION_PARAMETERS, int bc)
 	}
 
 	if (connect) {
-		php_mongo_connect(link TSRMLS_CC);
+		php_mongo_connect(link, MONGO_CON_FLAG_READ|MONGO_CON_FLAG_DONT_FILTER TSRMLS_CC);
 	}
 }
 /* }}} */
@@ -468,7 +500,7 @@ PHP_METHOD(MongoClient, connect)
 	mongoclient *link;
 
 	PHP_MONGO_GET_LINK(getThis());
-	RETURN_BOOL(php_mongo_connect(link TSRMLS_CC) != NULL);
+	RETURN_BOOL(php_mongo_connect(link, MONGO_CON_FLAG_READ TSRMLS_CC) != NULL);
 }
 /* }}} */
 
@@ -499,7 +531,7 @@ PHP_METHOD(MongoClient, close)
 			while (ptr) {
 				current = ptr;
 				ptr = ptr->next;
-				close_connection(link->manager, current->connection);
+				close_connection(link->manager, (mongo_connection *)current->data);
 				count++;
 			}
 
@@ -524,7 +556,7 @@ PHP_METHOD(MongoClient, close)
 	if (error_message) {
 		free(error_message);
 	}
-  RETURN_TRUE;
+	RETURN_TRUE;
 }
 /* }}} */
 
@@ -581,6 +613,7 @@ PHP_METHOD(MongoClient, selectDB)
 	char *db;
 	int db_len;
 	mongoclient *link;
+	int free_this_ptr = 0;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &db, &db_len) == FAILURE) {
 		return;
@@ -607,6 +640,7 @@ PHP_METHOD(MongoClient, selectDB)
 		if (link->servers->server[0]->username && link->servers->server[0]->password) {
 			zval       *new_link;
 			mongoclient *tmp_link;
+			int i;
 		
 			if (strcmp(link->servers->server[0]->db, "admin") == 0) {
 				mongo_manager_log(
@@ -627,9 +661,16 @@ PHP_METHOD(MongoClient, selectDB)
 
 				tmp_link->manager = link->manager;
 				tmp_link->servers = calloc(1, sizeof(mongo_servers));
-				mongo_servers_copy(tmp_link->servers, link->servers, MONGO_SERVER_COPY_NONE);
+				mongo_servers_copy(tmp_link->servers, link->servers, MONGO_SERVER_COPY_CREDENTIALS);
+				/* We assume the previous credentials will work on this
+				 * database too, or if authSource is set, authenticate against
+				 * that database */
+				for (i = 0; i < tmp_link->servers->count; i++) {
+					tmp_link->servers->server[i]->db = strdup(db);
+				}
 
 				this_ptr = new_link;
+				free_this_ptr = 1;
 			}
 		}
 	}
@@ -638,6 +679,9 @@ PHP_METHOD(MongoClient, selectDB)
 	MONGO_METHOD2(MongoDB, __construct, &temp, return_value, getThis(), name);
 
 	zval_ptr_dtor(&name);
+	if (free_this_ptr) {
+		zval_ptr_dtor(&this_ptr);
+	}
 }
 /* }}} */
 
@@ -674,7 +718,7 @@ PHP_METHOD(MongoClient, selectCollection)
 	zval *db_name, *coll_name, *temp_db;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ss", &db, &db_len, &coll, &coll_len) == FAILURE) {
-	return;
+		return;
 	}
 
 	MAKE_STD_ZVAL(db_name);
@@ -760,38 +804,37 @@ PHP_METHOD(MongoClient, dropDB)
 }
 /* }}} */
 
-
 /* {{{ MongoClient->listDBs
  */
-PHP_METHOD(MongoClient, listDBs) {
-  zval *admin, *data, *db;
+PHP_METHOD(MongoClient, listDBs)
+{
+	zval *admin, *data, *db;
 
-  MAKE_STD_ZVAL(admin);
-  ZVAL_STRING(admin, "admin", 1);
+	MAKE_STD_ZVAL(admin);
+	ZVAL_STRING(admin, "admin", 1);
 
-  MAKE_STD_ZVAL(db);
+	MAKE_STD_ZVAL(db);
 
-  MONGO_METHOD1(MongoClient, selectDB, db, getThis(), admin);
+	MONGO_METHOD1(MongoClient, selectDB, db, getThis(), admin);
 
-  zval_ptr_dtor(&admin);
+	zval_ptr_dtor(&admin);
 
-  MAKE_STD_ZVAL(data);
-  array_init(data);
-  add_assoc_long(data, "listDatabases", 1);
+	MAKE_STD_ZVAL(data);
+	array_init(data);
+	add_assoc_long(data, "listDatabases", 1);
 
-  MONGO_CMD(return_value, db);
+	MONGO_CMD(return_value, db);
 
-  zval_ptr_dtor(&data);
-  zval_ptr_dtor(&db);
+	zval_ptr_dtor(&data);
+	zval_ptr_dtor(&db);
 }
 /* }}} */
 
-
 /* {{{ MongoClient->getHosts
  */
 PHP_METHOD(MongoClient, getHosts)
 {
-	mongoclient             *link;
+	mongoclient            *link;
 	mongo_con_manager_item *item;
 
 	PHP_MONGO_GET_LINK(getThis());
@@ -803,27 +846,27 @@ PHP_METHOD(MongoClient, getHosts)
 		zval *infoz;
 		char *host;
 		int   port;
+		mongo_connection *con = (mongo_connection*) item->data;
 
 		MAKE_STD_ZVAL(infoz);
 		array_init(infoz);
 
-		mongo_server_split_hash(item->connection->hash, (char**) &host, (int*) &port, NULL, NULL, NULL, NULL, NULL);
+		mongo_server_split_hash(con->hash, (char**) &host, (int*) &port, NULL, NULL, NULL, NULL, NULL);
 		add_assoc_string(infoz, "host", host, 1);
 		add_assoc_long(infoz, "port", port);
 		free(host);
 
 		add_assoc_long(infoz, "health", 1);
-		add_assoc_long(infoz, "state", item->connection->connection_type == MONGO_NODE_PRIMARY ? 1 : (item->connection->connection_type == MONGO_NODE_SECONDARY ? 2 : 0));
-		add_assoc_long(infoz, "ping", item->connection->ping_ms);
-		add_assoc_long(infoz, "lastPing", item->connection->last_ping);
+		add_assoc_long(infoz, "state", con->connection_type == MONGO_NODE_PRIMARY ? 1 : (con->connection_type == MONGO_NODE_SECONDARY ? 2 : 0));
+		add_assoc_long(infoz, "ping", con->ping_ms);
+		add_assoc_long(infoz, "lastPing", con->last_ping);
 
-		add_assoc_zval(return_value, item->connection->hash, infoz);
+		add_assoc_zval(return_value, con->hash, infoz);
 		item = item->next;
 	}
 }
 /* }}} */
 
-
 /* {{{ proto static array Mongo::getConnections(void)
    Returns an array of all open connections, and information about each of the servers */
 PHP_METHOD(MongoClient, getConnections)
@@ -841,6 +884,7 @@ PHP_METHOD(MongoClient, getConnections)
 		zval *entry, *server, *connection, *tags;
 		char *host, *repl_set_name, *database, *username, *auth_hash;
 		int port, pid, i;
+		mongo_connection *con = (mongo_connection*) ptr->data;
 
 		MAKE_STD_ZVAL(entry);
 		array_init(entry);
@@ -855,7 +899,7 @@ PHP_METHOD(MongoClient, getConnections)
 		array_init(tags);
 
 		/* Grab server information */
-		mongo_server_split_hash(ptr->connection->hash, &host, &port, &repl_set_name, &database, &username, &auth_hash, &pid);
+		mongo_server_split_hash(con->hash, &host, &port, &repl_set_name, &database, &username, &auth_hash, &pid);
 
 		add_assoc_string(server, "host", host, 1);
 		free(host);
@@ -879,20 +923,20 @@ PHP_METHOD(MongoClient, getConnections)
 		add_assoc_long(server, "pid", pid);
 
 		/* Grab connection info */
-		add_assoc_long(connection, "last_ping", ptr->connection->last_ping);
-		add_assoc_long(connection, "last_ismaster", ptr->connection->last_ismaster);
-		add_assoc_long(connection, "ping_ms", ptr->connection->ping_ms);
-		add_assoc_long(connection, "connection_type", ptr->connection->connection_type);
-		add_assoc_string(connection, "connection_type_desc", mongo_connection_type(ptr->connection->connection_type), 1);
-		add_assoc_long(connection, "max_bson_size", ptr->connection->max_bson_size);
-		add_assoc_long(connection, "tag_count", ptr->connection->tag_count);
-		for (i = 0; i < ptr->connection->tag_count; i++) {
-			add_next_index_string(tags, ptr->connection->tags[i], 1);
+		add_assoc_long(connection, "last_ping", con->last_ping);
+		add_assoc_long(connection, "last_ismaster", con->last_ismaster);
+		add_assoc_long(connection, "ping_ms", con->ping_ms);
+		add_assoc_long(connection, "connection_type", con->connection_type);
+		add_assoc_string(connection, "connection_type_desc", mongo_connection_type(con->connection_type), 1);
+		add_assoc_long(connection, "max_bson_size", con->max_bson_size);
+		add_assoc_long(connection, "tag_count", con->tag_count);
+		for (i = 0; i < con->tag_count; i++) {
+			add_next_index_string(tags, con->tags[i], 1);
 		}
 		add_assoc_zval(connection, "tags", tags);
 
 		/* Top level elements */
-		add_assoc_string(entry, "hash", ptr->connection->hash, 1);
+		add_assoc_string(entry, "hash", con->hash, 1);
 		add_assoc_zval(entry, "server", server);
 		add_assoc_zval(entry, "connection", connection);
 		add_next_index_zval(return_value, entry);
@@ -901,3 +945,12 @@ PHP_METHOD(MongoClient, getConnections)
 	}
 }
 /* }}} */
+
+/*
+ * Local variables:
+ * tab-width: 4
+ * c-basic-offset: 4
+ * End:
+ * vim600: fdm=marker
+ * vim: noet sw=4 ts=4
+ */
